#include "stdafx.h"

#include <Compiler.h>


// f[^e[uXPW[
void Linker::ResolveDataTableSchedules( long dataSectionBaseOffset )
{
	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::DataTable )
		{
			nativeCode.Overwrite(
				schedule.GetOffset(),
				static_cast<long>( nativeCode.GetLong( schedule.GetOffset() ) + imageBase + dataSectionBaseOffset )
			);
		}
	}

	foreach( const Schedule &schedule, dataTable.schedules )
	{
		if( schedule.GetType() == Schedule::DataTable )
		{
#if defined _WIN64 || defined BUILD_X64_COMPILER
			dataTable.OverwriteInt64(
				schedule.GetOffset(),
				dataTable.GetInt64( schedule.GetOffset() ) + imageBase + dataSectionBaseOffset
			);
#else
			dataTable.Overwrite(
				schedule.GetOffset(),
				dataTable.GetLong( schedule.GetOffset() ) + imageBase + dataSectionBaseOffset
			);
#endif
		}
	}
}

// CatchAhX XPW[
void Linker::ResolveCatchAddressSchedules( long codeSectionBaseOffset )
{
	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::CatchAddress )
		{
			if( nativeCode.GetLong( schedule.GetOffset() ) != 0 )
			{
				// ul0̏ꍇ
				nativeCode.Overwrite(
					schedule.GetOffset(),
					static_cast<long>( nativeCode.GetLong( schedule.GetOffset() ) + schedule.GetUserProc().GetBeginOpAddress() + imageBase + codeSectionBaseOffset )
				);
			}
		}
	}

	foreach( const Schedule &schedule, dataTable.schedules )
	{
		if( schedule.GetType() == Schedule::CatchAddress )
		{
			if( dataTable.GetLong( schedule.GetOffset() ) != 0 )
			{
				// ul0̏ꍇ
#if defined _WIN64 || defined BUILD_X64_COMPILER
				dataTable.OverwriteInt64(
					schedule.GetOffset(),
					dataTable.GetInt64( schedule.GetOffset() ) + schedule.GetUserProc().GetBeginOpAddress() + imageBase + codeSectionBaseOffset
				);
#else
				dataTable.Overwrite(
					schedule.GetOffset(),
					dataTable.GetLong( schedule.GetOffset() ) + schedule.GetUserProc().GetBeginOpAddress() + imageBase + codeSectionBaseOffset
				);
#endif
			}
		}
	}
}

// DLL֐XPW[
void Linker::ResolveDllProcSchedules( long codeSectionBaseOffset, long importSectionBaseOffset, long lookupSize, long hintSize )
{
	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::DllProc )
		{
#if defined _AMD64_ || defined BUILD_X64_COMPILER
			nativeCode.Overwrite(
				schedule.GetOffset(),
				static_cast<long>( importSectionBaseOffset + schedule.GetDllProc().GetLookupAddress()
					- ( codeSectionBaseOffset + schedule.GetOffset() + sizeof(long) ) )
			);
#else
			nativeCode.Overwrite(
				schedule.GetOffset(),
				static_cast<long>( imageBase + importSectionBaseOffset + lookupSize + hintSize
					+ schedule.GetDllProc().GetLookupAddress() )
			);
#endif
		}
	}
}

// [U`֐XPW[
void Linker::ResolveUserProcSchedules( long codeSectionBaseOffset )
{
	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::UserProc
			|| schedule.GetType() == Schedule::AddressOf )
		{
			if( schedule.GetUserProc().GetBeginOpAddress() == 0
				&& schedule.GetUserProc().GetEndOpAddress() == 0 )
			{
				compiler.errorMessenger.OutputFatalError();
			}

			if( schedule.GetType() == Schedule::UserProc )
			{
				nativeCode.Overwrite(
					schedule.GetOffset(),
					static_cast<long>( schedule.GetUserProc().GetBeginOpAddress() - ( schedule.GetOffset() + sizeof(long) ) )
				);
			}
			else if( schedule.GetType() == Schedule::AddressOf )
			{
				nativeCode.Overwrite(
					schedule.GetOffset(),
					static_cast<long>( schedule.GetUserProc().GetBeginOpAddress() + imageBase + codeSectionBaseOffset )
				);
			}
		}
	}
}

// O[oϐXPW[
void Linker::ResolveGlobalVarSchedules( long rwSectionBaseOffset )
{
	int allInitVarSize = compiler.GetObjectModule().meta.GetGlobalVars().initAreaBuffer.GetSize();

	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::GlobalVar )
		{
			if( nativeCode.GetLong( schedule.GetOffset() ) & 0x80000000 )
			{
				nativeCode.Overwrite(
					schedule.GetOffset(),
					static_cast<long>( allInitVarSize + (nativeCode.GetLong( schedule.GetOffset() ) & 0x7FFFFFFF) + imageBase + rwSectionBaseOffset )
				);
			}
			else
			{
				nativeCode.Overwrite(
					schedule.GetOffset(),
					static_cast<long>( nativeCode.GetLong( schedule.GetOffset() ) + imageBase + rwSectionBaseOffset )
				);
			}
		}
	}
}

void Linker::ResolveVtblSchedule( long dataSectionBaseOffset )
{
	foreach( const Schedule &schedule, nativeCode.GetSchedules() )
	{
		if( schedule.GetType() == Schedule::ComVtbl )
		{
			LONG_PTR vtblOffset = schedule.GetClass().GetComVtblOffset();

			nativeCode.Overwrite(
				schedule.GetOffset(),
				static_cast<long>( vtblOffset + imageBase + dataSectionBaseOffset )
			);
		}

		if( schedule.GetType() == Schedule::Vtbl )
		{
			LONG_PTR vtblMasterListOffset = schedule.GetClass().GetVtblMasterListOffset();

			nativeCode.Overwrite(
				schedule.GetOffset(),
				static_cast<long>( vtblMasterListOffset + imageBase + dataSectionBaseOffset )
			);
		}
	}

	foreach( const Schedule &schedule, dataTable.schedules )
	{
		if( schedule.GetType() == Schedule::ComVtbl )
		{
			LONG_PTR vtblOffset = schedule.GetClass().GetComVtblOffset();

#if defined _WIN64 || defined BUILD_X64_COMPILER
			dataTable.OverwriteInt64(
				schedule.GetOffset(),
				vtblOffset + imageBase + dataSectionBaseOffset
			);
#else
			dataTable.Overwrite(
				schedule.GetOffset(),
				vtblOffset + imageBase + dataSectionBaseOffset
			);
#endif
		}

		if( schedule.GetType() == Schedule::Vtbl )
		{
			LONG_PTR vtblMasterListOffset = schedule.GetClass().GetVtblMasterListOffset();

#if defined _WIN64 || defined BUILD_X64_COMPILER
			dataTable.OverwriteInt64(
				schedule.GetOffset(),
				vtblMasterListOffset + imageBase + dataSectionBaseOffset
			);
#else
			dataTable.Overwrite(
				schedule.GetOffset(),
				vtblMasterListOffset + imageBase + dataSectionBaseOffset
			);
#endif
		}
	}
}

void Linker::ResolveTypeInfoSchedule( long dataSectionBaseOffset )
{
	foreach( const Schedule &schedule, dataTable.schedules )
	{
		if( schedule.GetType() == Schedule::TypeInfo )
		{
			LONG_PTR typeInfoDataTableOffset = schedule.GetClass().GetTypeInfoDataTableOffset();

#if defined _WIN64 || defined BUILD_X64_COMPILER
			dataTable.OverwriteInt64(
				schedule.GetOffset(),
				typeInfoDataTableOffset + imageBase + dataSectionBaseOffset
			);
#else
			dataTable.Overwrite(
				schedule.GetOffset(),
				typeInfoDataTableOffset + imageBase + dataSectionBaseOffset
			);
#endif
		}
	}
}

void Linker::Link( ObjectModule &masterObjectModule )
{
	// nativeCode͏ԂłȂ΂ȂȂ
	if( nativeCode.GetSize() > 0 )
	{
		compiler.errorMessenger.OutputFatalError();
	}

	nativeCode.PutEx( masterObjectModule.globalNativeCode );

	masterObjectModule.meta.GetUserProcs().Iterator_Reset();
	while( masterObjectModule.meta.GetUserProcs().Iterator_HasNext() )
	{
		const UserProc *pUserProc = masterObjectModule.meta.GetUserProcs().Iterator_GetNext();

		if( pUserProc->GetNativeCode().GetSize() > 0 )
		{
			// ֐̊JnʒuiR[hZNṼItZbgj
			int beginCodePos = nativeCode.GetSize();

			// R[hZNVɊ֐̃lCeBuR[hǉ
			nativeCode.PutEx( pUserProc->GetNativeCode() );

			// ֐̏IʒuiR[hZNṼItZbgj
			int endCodePos = nativeCode.GetSize();

			// R[hZNV̊֐͈̔͂ݒ
			pUserProc->SetBeginOpAddress( beginCodePos );
			pUserProc->SetEndOpAddress( endCodePos );
		}
	}
}

void Linker::SetDataTable( DataTable &dataTable )
{
	this->dataTable.Add( dataTable );
}
